/* RECEIVE.C - Modified for VMS.
  This file is part of the NetFax system.

  (c) Copyright 1989 by David M. Siegel. 
      All rights reserved.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful, 
    but WITHOUT ANY WARRANTY; without even the implied warranty of 
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#ifdef VMS
#include <file.h>
#include <socket.h>
#else /* VMS */
#include <termios.h>
#include <fcntl.h>
#include <malloc.h>
#include <sys/param.h>
#include <sys/wait.h>
#endif /* VMS */

#include "../../lib/libutil/dispatch.h"
#include "../../lib/libutil/reap.h"
#include "../../lib/libfax/libfax.h"


static void print_fdcs_params(f, fp)
     FaxModem *f;
     FILE *fp;
{
    if (FAX_ISSET(f, FAX_F_FDCS)) {
	fprintf(fp, " vr=%s br=%s wd=%s ln=%s,ecm=%s",
		t30_vr_string(&f->fdcs_params),
		t30_br_string(&f->fdcs_params),
		t30_wd_string(&f->fdcs_params),
		t30_ln_string(&f->fdcs_params),
		t30_ec_string(&f->fdcs_params));
      }
}


int receive_pages(fmp, seq, info_fp)
     FaxModem *fmp;
     int	seq;
     FILE *info_fp;
{
    int page = 0;

    for (;;) {
	time_t start = time(0);
	char file[MAXPATHLEN];
	int fd;

	/*
	 * Tell the modem we are ready to go.
	 */
	switch (faxmodem_start_recv(fmp)) {
	  case RECV_OK:
	    break;
	  case RECV_FAILED:
	    fprintf(info_fp, "  page %d failed\n", page+1);
	    fprintf(info_fp, "\nReceived %d pages succesfully\n", page);
	    return (-1);
	  case RECV_DONE:
	    fprintf(info_fp, "\nReceived %d pages succesfully\n", page);
	    return (page);
	}

	/*
	 * Open the file to hold the G3 data.
	 */
	sprintf(file, "%s/%d_%d", INCOMING_QUEUE, seq, page++);
	if ((fd = open(file, O_WRONLY|O_CREAT, 0700)) < 0) {
	    log(L_ALERT, "output file open failed: %m");
	    fprintf(info_fp, "  page %d failed\n", page);
	    return (-1);
	}
	
	/*
	 * Read in the page.
	 */
	if (faxmodem_recv_page(fmp, fd) < 0) {
	    log(L_WARNING, "receive of page failed");
	    fprintf(info_fp, "  page %d failed\n", page);
	    close(fd);
	    return (-1);
	}

	/*
	 * All done with the page.
	 */
	log(L_INFO, "received page %d", page);
	fprintf(info_fp, "  page %d ok", page);
	print_fdcs_params(fmp, info_fp);
	fprintf(info_fp, " time=%d\n", time(0) - start);
	close(fd);
    }
}

/*
 * This function does the bulk of the work for receiving a new
 * fax.  It answers the modem, reads the fax, writes the G3 files,
 * and mails notification.
 */
handle_incoming_fax(fmp)
     FaxModem *fmp;
{
    time_t start = time(0);
    char seq_file[MAXPATHLEN];
    char info_file[MAXPATHLEN];
    FILE *info_fp;
    int seq;
    int rc;

    log(L_INFO, "handling an incoming fax");

    if (faxmodem_answer(fmp) < 0) {
	log(L_NOTICE, "can't anwser the phone");
	faxmodem_hangup(fmp);
	return (-1);
    }

    sprintf(seq_file, "%s/%s", INCOMING_QUEUE, SEQ_FILE);
    if ((seq = seq_next(seq_file)) < 0) {
	log(L_ALERT, "can't get incoming sequence number: %m");
	faxmodem_hangup(fmp);
	return (-1);
    }

    sprintf(info_file, "%s/FAX.LOG", INCOMING_QUEUE);
    if ((info_fp = fopen(info_file, "a+")) == NULL) {
	log(L_ALERT, "can't make info file: %m");
	faxmodem_hangup(fmp);
	return (-1);
    }

    /*
     * Print the header for the info file.
     */
    fprintf(info_fp, "Reception of fax %d started at %s", seq, ctime(&start));
    faxmodem_print_id_strings(fmp, info_fp);
    fprintf(info_fp, "\n");

    /*
     * Receive the pages.
     */
    rc = receive_pages(fmp, seq, info_fp);
    fprintf(info_fp, "\nTotal connect time was %d seconds.\n", 
	    time(0) - start);

    /*
     * All done.
     */
    faxmodem_hangup(fmp);
    fclose(info_fp);

    /*
     * Resync resync the modem, just in case.
     */
    if (faxmodem_sync(fmp, FAXMODEM_SYNC_TRIES) < 0)
      log(L_EMERG, "can't resync to the modem");

    return (rc);
}
